import requests
import json
from os.path import isfile
from tqdm import tqdm # completely optional; we might exclude it

############ Functions used to map IDs from one DB to another
##############################################################
# USAGE:
# generate_mappings(just_pdb_to_uniprot=False) | (dumps to file; checks version and updates automatically; ignores if current)
# PDB_Uniprot("P10970")                        | (checks mappings in the dump files, accepts PDB ID, PDB ID + " " + Chain, or Uniprot ID; no need to specify)


# Help functions
# Download file with replacament if already same filename locally
def download_file(url, output_file, headers=""):
    with open(output_file, 'w') as new_local_file:
            new_local_file.write(requests.get(url, headers=headers).content.decode('utf-8'))

def json_dump(input_object, output_file):
    try:
        with open(output_file, 'w') as outf:
            json.dump(input_object, outf)
        return True
    except:
        print("Something went wrong when dumping file")
        return False

def json_load(input_file):
    try:
        with open(input_file, 'r') as inpf:
            loaded = json.load(inpf)
        return loaded
    except:
        print("Couldn't load.")
        return False

##### PDB <=> Uniprot
# Check if local version of the mapping file is most recent and replace if not
# The returned value says whether the local file is up to date with the remote file
# None means: we don't know because we couldn't fint the online file.
# If no local file we return Fale.
# If you just want to check, best to set no_download to True
def PDB_Uniprot_update_list(local_list_file_path="pdb_chain_uniprot.lst", no_download=False):
    try:
        # Constant URL for periodically updated PDB<=>Uniprot mappings
        remote_url = "http://ftp.ebi.ac.uk/pub/databases/msd/sifts/text/pdb_chain_uniprot.lst"
        # Load first 100 bytes of the online file to extract its header
        remote_url_headers = {"Range": "bytes=0-100"}
        # Extracting the header (first line)
        remote_file_content = requests.get(remote_url, headers=remote_url_headers).content.decode('utf-8')
        remote_file_header = remote_file_content.split("\n")[0]
    except:
        print("Cannot establish connection. No remote file under this address.")
        return None

    # Check if local file exists
    try: 
        # If it does, extract local file's header
        with open(local_list_file_path, 'r') as local:
            local_file_header = local.readline().strip()
        
        # Compare the local file's header to the remote file's header
        if local_file_header == remote_file_header:
            print("The local mapping file is up to date.")
            return True
        else:
            print("There is a more up to date mapping file available online. Replacing it.")
            if not no_download:
                download_file(remote_url, output_file)
            return False

    except:
        # No file found. Downloading from remote.
        print("No local file found.")
        if not no_download:
            print("Downloading the remote file")
            download_file(remote_url, local_list_file_path)
        return False

# Parse mapping and put it into a dictionary with "PDBID Chain" key/value pattern
# PDB => Uniprot by default; set false to have it the other way round
def parse_mapping(local_list_file_path="pdb_chain_uniprot.lst", pdb_to_uniprot=True):
    try:
        with open(local_list_file_path, 'r') as raw_mapping:
            # Mapping as a list of lists, each line already divided into columns
            # Already without a header
            raw_mapping = [x.strip().split("\t") for x in raw_mapping.readlines()][1:]
    except:
        print("No local mapping file found. Download it first.")
        return False

    # Create a dictionary with the mapping, depending on the direction
    mapping = dict()
    if pdb_to_uniprot:
        print("Parsing the PDB => Uniprot mapping")
        for line in tqdm(raw_mapping):
            mapping[line[0] + " " + line[1]] = line[2]
    else:
        print("Parsing the Uniprot => PDB mapping")
        for line in tqdm(raw_mapping):
            if line[2] in mapping.keys():
                mapping[line[2]].append(line[0] + " " + line[1])
            else:
                mapping[line[2]] = [line[0] + " " + line[1]]
    
    return mapping

# Generates mappings if a more recent version can be found and saves them to file
# Returns True if generated anything, returns False if nothing has been generated
def generate_mappings(local_file="pdb_chain_uniprot.lst", just_pdb_to_uniprot=True, pdb_to_uni_mapping_path="pdb_to_uni_mapping.json", uni_to_pdb_mapping_path="uni_to_pdb_mapping.json"):
    # Check if up to date without downloading first
    if not PDB_Uniprot_update_list(local_list_file_path=local_file, no_download=True):
        # If not, download
        PDB_Uniprot_update_list(local_list_file_path=local_file)

        # Generate new mappings
        json_dump(parse_mapping(pdb_to_uniprot=True), pdb_to_uni_mapping_path)

        if not just_pdb_to_uniprot:
            json_dump(parse_mapping(pdb_to_uniprot=False), uni_to_pdb_mapping_path)
        
        return True

    # If mapping source file up to date but one of the generated mapping .json files is missing, generate new mapping
    elif not isfile(pdb_to_uni_mapping_path) or not isfile(uni_to_pdb_mapping_path):
        if not isfile(pdb_to_uni_mapping_path):
            json_dump(parse_mapping(pdb_to_uniprot=True), pdb_to_uni_mapping_path)
        
        if not isfile(uni_to_pdb_mapping_path) and not just_pdb_to_uniprot:
            json_dump(parse_mapping(pdb_to_uniprot=False), uni_to_pdb_mapping_path)
        
        return True

    else:
        print("Everything is up to date; no need to generate new mapping files.")
        return False

# Look up mapping PDB <=> Uniprot
# Search_key can be either a PDBid, a "PDBid Chain" pair or a Uniprot ID; no need to signal it to the function, it recognizes it automatically.
# If search_key == PDBid: return the list of all available mappings
# If search_key == PDBid + " " + Chain: return the only available mapping
# If search_key == UniprotID: return the list of all mappings
def PDB_Uniprot(search_key, pdb_to_uniprot_mapping_path="pdb_to_uni_mapping.json", uniprot_to_pdb_mapping_path="uni_to_pdb_mapping.json"):
    if isfile(pdb_to_uniprot_mapping_path):
        try:
            # load PDB => Uniprot mapping
            pdb_to_uni = json_load(pdb_to_uniprot_mapping_path)
        except:
            print("Something went wrong loading the PDB => Uniprot .json mapping file.")
            return False
    
    if isfile(uniprot_to_pdb_mapping_path):
        try:
            # load Uniprot => PDB mapping
            uni_to_pdb = json_load(uniprot_to_pdb_mapping_path)
        except:
            print("Something went wrong loading the Uniprot => PDB .json mapping file.")

    # If none could be loaded, abort
    if 'pdb_to_uni' not in locals() and 'uni_to_pdb' not in locals():
        print("No mapping could be loaded. Aborting lookup.")
        return False
    
    # If search_key == PDBid: return the list of all available mappings
    if len(search_key.split()) == 1 and len(search_key) == 4:
        if 'pdb_to_uni' in locals():
            matching_chains = [x for x in pdb_to_uni.keys() if x.startswith(search_key)]
            if len(matching_chains) > 0:
                matches = [tuple([matching_key, pdb_to_uni[matching_key]]) for matching_key in matching_chains]
                return matches
            else:
                print("No mapping found for PDB ID:", search_key)
                return False
        else:
            print("Missing PDB => Uniprot mapping.")
            return False
    
    # If search_key == PDBid + " " + Chain: return the only available mapping
    elif len(search_key.split()) == 2 and len(search_key.split()[0]) == 4:
        if 'pdb_to_uni' in locals():
            if search_key in pdb_to_uni.keys():
                return pdb_to_uni[search_key]
            else:
                print("No mapping found for:", search_key)
                return False
        else:
            print("Missing PDB => Uniprot mapping.")
            return False
    
    # If search_key == UniprotID: return the list of all mappings
    elif len(search_key.split()) == 1 and len(search_key) > 4:
        if 'uni_to_pdb' in locals():
            if search_key in uni_to_pdb.keys():
                return uni_to_pdb[search_key]
            else:
                print("No mapping found for Uniprot ID:", search_key)
                return False
        else:
            print("Missing Uniprot => PDB mapping.")
            return False
    
    # Incorrect search_key
    else:
        print("Incorrect search phrase. Are you sure you're trying to map PDB <=> Uniprot?\nLegal search phrases: 1s72, 1s72 A, P30615")
        return False


from fetchPDBinfo import fetchPDBinfo
from getAnnotations import getPfamAnnotation, getAnnotations, getECAnnotation


def test_get_annotations():
    a = getAnnotations("1j85", "A")
    assert str(a.get()[0]) =="['1j85', 'A', '2', '142', 'PF00588.18', 'SpoU_methylase', 'SpoU rRNA Methylase family', '4.3E-32']"


def test_get_ec_annotations():
    a = getECAnnotation("1cak", "A")
    assert str(a.get()) == "[('4.2.1.1', 'Carbonic anhydrase.')]"


def test_get_pfam_annotations():
    a = getPfamAnnotation("1j85", "A")
    assert str(a.get()) == "[('1j85', 'A', 'SpoU_methylase', 'SpoU rRNA Methylase family', 'PF00588')]"


def test_fetch_pdb_info():
    a = fetchPDBinfo(".", "1uak", "A")
    assert str(a.getCalfaBreaks()) == '[162, 172]'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Pawel Rubach 2019
# based on:
# Copyright Michal Jamroz, 2014, jamroz@chem.uw.edu.pl
import datetime
import urllib.request, urllib.error, urllib.parse,gzip,re 
import os
from os import path
from lxml import etree


REFRESH_FILE_INTERVAL = 15*24*3600 # Refresh every week

PDB_CHAIN_ENZYME = [ "http://ftp.ebi.ac.uk/pub/databases/msd/sifts/flatfiles/tsv/pdb_chain_enzyme.tsv.gz", "pdb_chain_enzyme.tsv.gz" ]
ENZYME_DAT = [ "http://ftp.expasy.org/databases/enzyme/enzyme.dat", "enzyme.dat"]

PDB_CHAIN_PFAM = [ "http://ftp.ebi.ac.uk/pub/databases/msd/sifts/flatfiles/tsv/pdb_chain_pfam.tsv.gz", "pdb_chain_pfam.tsv.gz" ]
PFAM_DESC = [ "http://pfam.xfam.org/families?output=text", "pdb_chain_pfam_desc" ]

DEFAULT_DATA_FILE_PATH = path.join(path.expanduser("~"), ".local","spyprot")


class AnnotationBase:
    '''
       Base class for retrieving annotations for a given PDBID and Chain

       Annotations are parsed from files downloaded from databases and stored locally in:
       $HOME/.local/spyprot/
    '''
    def __init__(self, files, pdb, chain='A', data_file_path=DEFAULT_DATA_FILE_PATH, refresh_file_interval=REFRESH_FILE_INTERVAL):
        self.data_file_path = data_file_path
        if not path.exists(data_file_path):
            os.makedirs(data_file_path)
        self.refresh_file_interval = refresh_file_interval
        self.first_file = self.download_if_not_exist(files[0])
        self.sec_file = self.download_if_not_exist(files[1])
        self.pdb = pdb.lower()
        self.chain = chain
        self.results = []
        self.d = []
        self.identifiers = []

    def get(self):
        return self.results

    @staticmethod
    def touch(path):
        with open(path, 'a'):
            os.utime(path, None)

    def download_if_not_exist(self, file_locs):
        enz_file = path.join(self.data_file_path, file_locs[1])
        if not path.isfile(enz_file) or (not path.isfile(enz_file + ".flg") and path.isfile(enz_file) and datetime.datetime.now().timestamp()-os.stat(enz_file).st_mtime > self.refresh_file_interval):
            print('downloading file: ' + enz_file + " from: "  + file_locs[0])
            AnnotationBase.touch(enz_file + ".flg")
            try:
                f = urllib.request.urlopen(file_locs[0])
                fw = open(enz_file + "_NEW", "wb")
                fw.write(f.read())
                fw.close()
                if path.isfile(enz_file):
                    os.rename(enz_file, enz_file + "_OLD")
                os.rename(enz_file + "_NEW", enz_file)
                if path.isfile(enz_file + "_OLD"):
                    os.remove(enz_file + "_OLD")
            except Exception as e:
                print("Problem updating file: " + enz_file + " leaving old one" + str(e))
            if path.isfile(enz_file + ".flg"): os.remove(enz_file + ".flg")
        return enz_file


class getECAnnotation(AnnotationBase):
    '''Retrieve EC annotations for a given PDBID and Chain

       Annotations are parsed from files downloaded from expasy and stored locally in:
       $HOME/.local/spyprot/

       Parameters
       ==========
       pdbcode: string
       chain: string

       Return
       ======
       results: list of tuples of EC Code and EC name
       '''
    def __init__(self, pdb, chain='A'):
        super().__init__([PDB_CHAIN_ENZYME, ENZYME_DAT], pdb, chain)
        q = re.compile(r'^'+self.pdb+'\t'+self.chain+'\t(.*)\t(?P<ec>.*)$')
        with gzip.open(self.first_file) as s:
            for line in s.readlines():
                z = q.match(line.decode('utf-8'))
                if z:
                    self.identifiers.append( z.group('ec') )
            for e in self.identifiers:
                o = self._getNameLocal(e)
                if o == "" and not e.endswith("-"):
                    o = self._getName(e)
                self.results.append( (e, o) )

    def _getName(self,ec):
        o = ""
        try:
            f = urllib.request.urlopen("https://enzyme.expasy.org/EC/"+ec+".txt")
            q = re.compile(r'^DE.{3}(?P<name>.*)$')
            for line in f.readlines():
                z = q.match(line.decode('utf-8'))
                if z:
                    o = z.group('name')
                    break
                
        except Exception as e:
            print("Problem getting EC name: " + ec + "\n" + str(e))
        return o

    def _getNameLocal(self, ec):
        if ec.endswith("-"):
            return ""
        o = ""
        try:
            with open(self.sec_file, encoding='utf-8') as f:
                q = re.compile(r'^ID.{3}(' + ec + ')$')
                id_fnd = False
                for line in f.readlines():
                    z = q.match(line)
                    if z and not id_fnd:
                        q = re.compile(r'^DE.{3}(?P<name>.*)$')
                        id_fnd = True
                        continue
                    if z and id_fnd:
                        o = z.group('name')
                        break

        except Exception as e:
            print("Problem getting EC name: " + ec + "\n" + str(e))
        return o


class getPfamAnnotation(AnnotationBase):
    '''Retrieve PFAM annotations for a given PDBID and Chain

       Annotations are parsed from files downloaded from pfam and stored locally in:
       $HOME/.local/spyprot/

       Parameters
       ==========
       pdbcode: string
       chain: string

       Return
       ======
       results: list of tuples containing attributes: pdbid, chain, pfam_short, pfam_desc and pfam accession code
       '''
    def __init__(self,pdb,chain='A'):
        super().__init__([PDB_CHAIN_PFAM, PFAM_DESC], pdb, chain)
        q = re.compile(r'^'+self.pdb+'\t'+self.chain+'\t(.*)\t(?P<pfam>.*)\t(.*)$', re.M)
        with gzip.open(self.first_file) as s:
            for line in s.readlines():
                line = line.decode('utf-8')
                z = q.match(line)
                if z:
                    self.identifiers.append( z.group('pfam') )
            for e in self.identifiers:
                a, b, c = self._getNames(e)
                self.results.append((pdb, chain, a, b, c))

    def _getNames(self, pfam):
        q = re.compile(r'^'+pfam+'\t(?P<pfam_short>.*)\t(?P<pfam_desc>.*)$')
        with open(self.sec_file, encoding='utf-8') as sd:
            for line in sd:
                z = q.match(line)
                if z:
                    return z.group('pfam_short'),z.group('pfam_desc'),pfam
        return "","",""


class getAnnotations:
    '''Retrieve PFAM annotations for a given PDBID and Chain directly from RCSB

       Parameters
       ==========
       pdbcode: string
       chain: string

       Return
       ======
       results: list of tuples containing attributes: pdbid, chain, ..., ..., pfam accession code, pfam_short, pfam_desc, ...
       '''
    def __init__(self, pdb, chain='A'):
        self.pdb = pdb
        self.chain = chain
        url = "http://www.pdb.org/pdb/rest/hmmer?structureId="+pdb.upper()
        r = urllib.request.urlopen(url)
        data = r.read()
        self.hmmer = etree.fromstring(data)

    def get(self):
        out=[]
        for r in self.hmmer:
            d = list(r.values())
            d[0] = d[0].lower()
            out.append(d)
        #out[0] = out[0].lower()
        return out


if __name__ == '__main__':
    a = getPfamAnnotation("1j85", "A")
    print(str(a.get()))
    a = getAnnotations("1j85", "A")
    print(str(a.get()))
    a = getECAnnotation("1cak", "A")
    print(str(a.get()))


#!/usr/bin/env python
import urllib.error
import urllib.parse
import urllib.request
from itertools import count, groupby
from os import path

from lxml import etree


def convertToRanges(L):
    G=(list(x) for _,x in groupby(L, lambda x,c=count(): next(c)-x))
    return ["-".join(map(str,(g[0],g[-1])[:len(g)])) for g in G]


class XmlPdbParser:
    def __init__(self, work_dir, pdbcode, chain, localfile=False):
        self.pdb = pdbcode
        self.chain = chain
        self.work_dir = work_dir
        self.localfile = localfile
        self.xmlgzfile = self.download_if_not_exist()
        self.root = etree.parse(self.xmlgzfile, etree.XMLParser(ns_clean=True)).getroot()
        self.NS = self.root.nsmap

    def download_if_not_exist(self):
        pdb_xml_file = path.join(self.work_dir, self.pdb + '.xml.gz')
        if not path.isfile(pdb_xml_file):
            f = urllib.request.urlopen('http://www.rcsb.org/pdb/files/' + self.pdb + '.xml.gz')
            fw = open(pdb_xml_file, "wb")
            fw.write(f.read())
            fw.close()
            f.close()
        return pdb_xml_file


class getCoordinates(XmlPdbParser):
    def __init__(self, work_dir, pdbcode, localfile=False):
        super().__init__(work_dir, pdbcode, None, localfile)
        l = self.root.xpath("//PDBx:atom_siteCategory/PDBx:atom_site[PDBx:auth_atom_id='CA'][PDBx:pdbx_PDB_model_num='1'][PDBx:group_PDB='ATOM']/PDBx:auth_asym_id",namespaces=self.NS)
        self.chains = set()
        self.chains.update(e.text for e in l)

    def getChainIndexes(self):
        return self.chains

    def getCalfa(self,chain,output='',preserve_seqid=False):
        l = self.root.xpath("//PDBx:atom_siteCategory/PDBx:atom_site[not(PDBx:label_alt_id/text()) or PDBx:label_alt_id='A'][PDBx:auth_atom_id='CA'][PDBx:pdbx_PDB_model_num='1'][PDBx:group_PDB='ATOM' or (PDBx:group_PDB='HETATM' and PDBx:label_comp_id='MSE') or (PDBx:group_PDB='HETATM' and PDBx:label_comp_id='ORN') or (PDBx:group_PDB='HETATM' and PDBx:label_comp_id='PCA') or (PDBx:group_PDB='HETATM' and PDBx:label_comp_id='DGL')][PDBx:auth_asym_id='"+chain+"']",namespaces=self.NS) 
# mozna by pewnie zrobic inaczej or hetatom and (xxx or xxx or xxx ....)
        o = ''
        #r = []
        prev_seqid=999999999999999
        first_resid = -1
        for e in l:
            d = e.find("PDBx:label_seq_id",namespaces=self.NS)
            if d.text:
                first_resid = int(d.text)
                break
        prev_seqid = first_resid-9999999999999     
        for e in l:
            x = e.find("PDBx:Cartn_x",namespaces=self.NS)
            y = e.find("PDBx:Cartn_y",namespaces=self.NS)
            z = e.find("PDBx:Cartn_z",namespaces=self.NS)
            seqid = e.find("PDBx:label_seq_id",namespaces=self.NS)
            if not seqid.text:
                continue
            if int(seqid.text)==prev_seqid:
                prev_seqid = int(seqid.text)
                print(prev_seqid)
                continue
            prev_seqid = int(seqid.text)
            l = [seqid,x,y,z]
            if None not in l:
                if preserve_seqid:
                    new_seqid = int(l[0].text)
                else:
                    new_seqid = int(l[0].text)-first_resid+1
                o+= "%4d %8.3f %8.3f %8.3f\n"% (new_seqid, float(l[1].text), float(l[2].text), float(l[3].text) )
        if output !='':
            f = open(output,"w")
            f.write(o)
            f.close()
        else:
            return o

    def getCalfaPdbFormat(self,chain,output='',preserve_seqid=False):
        l = self.root.xpath("//PDBx:atom_siteCategory/PDBx:atom_site[not(PDBx:label_alt_id/text()) or PDBx:label_alt_id='A'][PDBx:auth_atom_id='CA'][PDBx:pdbx_PDB_model_num='1'][PDBx:group_PDB='ATOM' or (PDBx:group_PDB='HETATM' and PDBx:label_comp_id='MSE') or (PDBx:group_PDB='HETATM' and PDBx:label_comp_id='ORN') or (PDBx:group_PDB='HETATM' and PDBx:label_comp_id='PCA') or (PDBx:group_PDB='HETATM' and PDBx:label_comp_id='DGL')][PDBx:auth_asym_id='"+chain+"']",namespaces=self.NS) 
        o = ''
        r = []
        prev_seqid=999999999999999
        first_resid = -1
        for e in l:
            d = e.find("PDBx:label_seq_id",namespaces=self.NS)
            if d.text:
                first_resid = int(d.text)
                break
        prev_seqid = first_resid-9999999999999     
                
        for e in l:
            x = e.find("PDBx:Cartn_x",namespaces=self.NS)
            y = e.find("PDBx:Cartn_y",namespaces=self.NS)
            z = e.find("PDBx:Cartn_z",namespaces=self.NS)
            seqname = e.find("PDBx:label_comp_id",namespaces=self.NS)
            seqid = e.find("PDBx:label_seq_id",namespaces=self.NS)
            true_seqid = e.find("PDBx:auth_seq_id",namespaces=self.NS)
            if not seqid.text:
                continue
            if preserve_seqid and not true_seqid.text:
                continue
            if int(seqid.text)==prev_seqid:
                prev_seqid = int(seqid.text)
                continue
            prev_seqid = int(seqid.text)
            bfactor = e.find("PDBx:B_iso_or_equiv",namespaces=self.NS) 
            l = (seqid,seqname,chain,seqid,x,y,z,bfactor)
            if None not in l:
                if preserve_seqid:
                    new_seqid = int(true_seqid.text)
                else:
                    new_seqid = int(seqid.text)-first_resid+1
                l = (new_seqid,str(seqname.text),str(chain),new_seqid,float(x.text),float(y.text),float(z.text),float(bfactor.text))
                o+= "ATOM%7d  CA%5s%2s%4d%12.3f%8.3f%8.3f  1.00%6.2f           C\n"%l
        if output !='':
            f = open(output,"w")
            f.write(o)
            f.close()
        else:
            return o


class fetchPDBinfo(XmlPdbParser):
    def __init__(self, work_dir, pdbcode, chain='A', localfile=False):
        super().__init__(work_dir, pdbcode, chain, localfile)
        self.codification = { "ALA" : 'A',
                 "CYS" : 'C',
                 "ASP" : 'D',
                 "GLU" : 'E',
                 "PHE" : 'F',
                 "GLY" : 'G',
                 "HIS" : 'H',
                 "ILE" : 'I',
                 "LYS" : 'K',
                 "LEU" : 'L',
                 "MET" : 'M',
                 "MSE" : 'M',
                 "ASN" : 'N',
                 "PYL" : 'O',
                 "PRO" : 'P',
                 "GLN" : 'Q',
                 "ARG" : 'R',
                 "SER" : 'S',
                 "THR" : 'T',
                 "SEC" : 'U',
                 "VAL" : 'V',
                 "TRP" : 'W',
                 "5HP" : 'E',
                 "ABA" : 'A',
                 "AIB" : 'A',
                 "BMT" : 'T',
                 "CEA" : 'C',
                 "CGU" : 'E',
                 "CME" : 'C',
                 "CRO" : 'X',
                 "CSD" : 'C',
                 "CSO" : 'C',
                 "CSS" : 'C',
                 "CSW" : 'C',
                 "CSX" : 'C',
                 "CXM" : 'M',
                 "DAL" : 'A',
                 "DAR" : 'R',
                 "DCY" : 'C',
                 "DGL" : 'E',
                 "DGN" : 'Q',
                 "DHI" : 'H',
                 "DIL" : 'I',
                 "DIV" : 'V',
                 "DLE" : 'L',
                 "DLY" : 'K',
                 "DPN" : 'F',
                 "DPR" : 'P',
                 "DSG" : 'N',
                 "DSN" : 'S',
                 "DSP" : 'D',
                 "DTH" : 'T',
                 "DTR" : 'X',
                 "DTY" : 'Y',
                 "DVA" : 'V',
                 "FME" : 'M',
                 "HYP" : 'P',
                 "KCX" : 'K',
                 "LLP" : 'K',
                 "MLE" : 'L',
                 "MVA" : 'V',
                 "NLE" : 'L',
                 "OCS" : 'C',
                 "ORN" : 'A',
                 "PCA" : 'E',
                 "PTR" : 'Y',
                 "SAR" : 'G',
                 "SEP" : 'S',
                 "STY" : 'Y',
                 "TPO" : 'T',
                 "TPQ" : 'F',
                 "TYS" : 'Y',
                 "TYR" : 'Y' }
        self._getCAmissing()

    def getCalfaBreaks(self,preserve_seqid=False):
        l = self.root.xpath("//PDBx:atom_siteCategory/PDBx:atom_site[not(PDBx:label_alt_id/text()) or PDBx:label_alt_id='A'][PDBx:auth_atom_id='CA'][PDBx:pdbx_PDB_model_num='1'][PDBx:group_PDB='ATOM' or (PDBx:group_PDB='HETATM' and PDBx:label_comp_id='MSE') or (PDBx:group_PDB='HETATM' and PDBx:label_comp_id='ORN') or (PDBx:group_PDB='HETATM' and PDBx:label_comp_id='PCA') or (PDBx:group_PDB='HETATM' and PDBx:label_comp_id='DGL')][PDBx:auth_asym_id='"+self.chain+"']",namespaces=self.NS) 
# mozna by pewnie zrobic inaczej or hetatom and (xxx or xxx or xxx ....)
        o = []
        #r = []
        prev_seqid=999999999999999
        first_resid = -1
        for e in l:
            d = e.find("PDBx:label_seq_id",namespaces=self.NS)
            if d is not None and d.text:
                first_resid = int(d.text)
                break
        prev_seqid = first_resid-9999999999999     
        for e in l:
            x = e.find("PDBx:Cartn_x",namespaces=self.NS)
            y = e.find("PDBx:Cartn_y",namespaces=self.NS)
            z = e.find("PDBx:Cartn_z",namespaces=self.NS)
            seqid = e.find("PDBx:label_seq_id",namespaces=self.NS)
            if not seqid.text:
                continue
            if int(seqid.text)==prev_seqid:
                prev_seqid = int(seqid.text)
                print(prev_seqid)
                continue
            prev_seqid = int(seqid.text)
            l = [seqid,x,y,z]
            if None not in l:
                if preserve_seqid:
                    new_seqid = int(l[0].text)
                else:
                    new_seqid = int(l[0].text)-first_resid+1
                o.append( (new_seqid, float(l[1].text), float(l[2].text), float(l[3].text) ) )
        # check chain breaks
        eps = 4.2*4.2
        brk = []
        for i in range(1,len(o)):
            i1 = o[i-1][0]
            i2 = o[i][0]
            x1 = o[i-1][1]
            y1 = o[i-1][2]
            z1 = o[i-1][3]

            x2 = o[i][1]
            y2 = o[i][2]
            z2 = o[i][3]
    
            d = (x1-x2)**2  + (y2-y1)**2 + (z1-z2)**2
            if d>eps:
                brk.append(i1)
                brk.append(i2)
        return brk
    def getFirstResidueIndex(self):
        return 1
    def getTrueFirstResidueIndex(self):
        l = self.root.xpath("//PDBx:atom_siteCategory/PDBx:atom_site[not(PDBx:label_alt_id/text()) or PDBx:label_alt_id='A'][PDBx:auth_atom_id='CA'][PDBx:pdbx_PDB_model_num='1'][PDBx:group_PDB='ATOM' or (PDBx:group_PDB='HETATM' and PDBx:label_comp_id='MSE') or (PDBx:group_PDB='HETATM' and PDBx:label_comp_id='ORN') or (PDBx:group_PDB='HETATM' and PDBx:label_comp_id='PCA') or (PDBx:group_PDB='HETATM' and PDBx:label_comp_id='DGL')][PDBx:auth_asym_id='"+self.chain+"']/PDBx:auth_seq_id",namespaces=self.NS) 
        for e in l:
            if e.text:
                #print e.text
                return int(e.text)
        return 1 # could be not true anymore
    def getSeqOneLetterCode(self):
        l = self.root.xpath("//PDBx:atom_siteCategory/PDBx:atom_site[PDBx:auth_atom_id='CA'][PDBx:auth_asym_id='"+self.chain+"'][PDBx:pdbx_PDB_model_num='1'][PDBx:group_PDB='ATOM' or (PDBx:group_PDB='HETATM' and PDBx:label_comp_id='MSE') or (PDBx:group_PDB='HETATM' and PDBx:label_comp_id='ORN') or (PDBx:group_PDB='HETATM' and PDBx:label_comp_id='PCA') or (PDBx:group_PDB='HETATM' and PDBx:label_comp_id='DGL')][not(PDBx:label_alt_id/text()) or PDBx:label_alt_id='A']/PDBx:label_comp_id",namespaces=self.NS) 
        k = list(self.codification.keys())
        seq = ""
        for e in l:
            s = e.text
            if s in k:
                seq += self.codification[s]
            else:
                seq += "X"
        return seq
    def getMissingArray(self):
        return self.missing_array
    def _getCAmissing(self):
        seq=[]
        l = self.root.xpath("//PDBx:atom_siteCategory/PDBx:atom_site[PDBx:auth_atom_id='CA'][PDBx:auth_asym_id='"+self.chain+"'][PDBx:pdbx_PDB_model_num='1'][PDBx:group_PDB='ATOM' or (PDBx:group_PDB='HETATM' and PDBx:label_comp_id='MSE') or (PDBx:group_PDB='HETATM' and PDBx:label_comp_id='ORN') or (PDBx:group_PDB='HETATM' and PDBx:label_comp_id='PCA') or (PDBx:group_PDB='HETATM' and PDBx:label_comp_id='DGL')][not(PDBx:label_alt_id/text()) or PDBx:label_alt_id='A']/PDBx:label_seq_id",
                            namespaces=self.NS)
        for e in l:
            if e.text: seq.append(int(e.text))
        if len(seq)>0:      #TODO make better function for getting missing - it fails e.g. in case of 3bpd_C
            first = seq[0]
        for e in range(len(seq)):
            seq[e] = seq[e] - first + 1 # przenumerowanie od 1 z zachowaniem przerw
        
        self.crystlen = len(seq)
        if len(seq)>0:
            self.missing = [x for x in range(seq[0],seq[-1]+1) if x not in seq]
            self.missing_array = self.missing
            self.missing = convertToRanges(self.missing)
            self.seq_idx = seq
            self.seq_len = seq[-1]
        else:
            self.missing=[]
            self.missing_array = self.missing
            self.seq_len = 0

    def getSeqLength(self):
        return self.seq_len
    def getPdbCreationDate(self):
        self.date = self.root.xpath("//PDBx:pdbx_database_statusCategory/PDBx:pdbx_database_status/PDBx:recvd_initial_deposition_date",namespaces=self.NS)[0].text
#        self.date = self.root.xpath("//PDBx:database_PDB_revCategory/PDBx:database_PDB_rev/PDBx:date_original",namespaces=self.NS)[0].text
        try:
            return self.date
        except:
            return None
    def getCAlen(self):
        return self.crystlen
    def getMissing(self):
        return self.missing


    def getPubtitlePubmed(self):
        pubmed = self.root.xpath("//PDBx:citationCategory/PDBx:citation[1]/PDBx:pdbx_database_id_PubMed",namespaces=self.NS)
        doi = self.root.xpath("//PDBx:citationCategory/PDBx:citation[1]/PDBx:pdbx_database_id_DOI",namespaces=self.NS)
        title = self.root.xpath("//PDBx:citationCategory/PDBx:citation[1]/PDBx:title",namespaces=self.NS)
        desc = self.root.xpath("//PDBx:structCategory/PDBx:struct/PDBx:title",namespaces=self.NS)
        src = self.root.xpath("//PDBx:entity_src_genCategory/PDBx:entity_src_gen/PDBx:pdbx_gene_src_scientific_name",namespaces=self.NS)
        key = self.root.xpath("//PDBx:struct_keywordsCategory/PDBx:struct_keywords/PDBx:pdbx_keywords",namespaces=self.NS)
        molecutag = self.root.xpath("//PDBx:entityCategory/PDBx:entity/PDBx:pdbx_description",namespaces=self.NS)

        doi = doi[0].text if doi else None
        molecutag = molecutag[0].text if molecutag else None
        key = key[0].text.capitalize() if key else None
        src = src[0].text if src else None
        pubmed = pubmed[0].text if pubmed else None
        title = title[0].text if title else None
        desc = desc[0].text.capitalize() if desc else None
        
        return (doi,pubmed,desc,title,src,key,molecutag)

if __name__ == "__main__":
#    pdb=argv[1].lower()
    b = getCoordinates("/tmp", "1uak")
    b.getCalfaPdbFormat("A", output=path.join("/tmp", "t.pdb"))
    a = fetchPDBinfo("/tmp", "1uak","A")
    print(a.getCalfaBreaks())
    sequence = a.getSeqOneLetterCode()
    print("seq",sequence)
    str_len = int(a.getCAlen())
    print("str_len",str_len)
    seq_len = a.getSeqLength()
    print("seq_len",seq_len)
    missing = ", ".join(a.getMissing())
    print("miss",missing)
    date = a.getPdbCreationDate()
    print("date",date)
    structure_start = a.getFirstResidueIndex()
    print("start",structure_start)
    doi, pubmed, pubtitle, prottitle, source, keyword, molname =  a.getPubtitlePubmed()
    print("rest","\n".join([doi, pubmed, pubtitle, prottitle, source, keyword, molname]))





#!/usr/bin/env python
# Copyright Michal Jamroz 2014, jamroz@chem.uw.edu.pl
import csv
import shutil
import tarfile
import urllib.request, urllib.error, urllib.parse
from os import makedirs, path
from xml.dom import minidom
from lxml import etree
from Bio.PDB import MMCIFIO, Select
from Bio.PDB.MMCIFParser import MMCIFParser


class ProteinFile:
    '''
       Base classs for download of PDB/CIF files from PDB Database and filter by chain and filter out HOH
       Supports also PDB Bundles when there are many subchains for a given protein
    '''
    def __init__(self, dir, pdbcode, chain=None):
        self.pdbcode = pdbcode
        self.chain = chain
        self.dir = dir


class ChainAndResidueSelect(Select):
    def __init__(self, chain, model=1, residue_out="HOH"):
        self.chain = chain
        self.model = model
        self.residue_out = residue_out

    def accept_model(self, model):
        if model is None or (model is not None and model.serial_num == self.model):
            return True
        else:
            return False

    def accept_chain(self, chain):
        if chain is None or (chain is not None and chain.id == self.chain):
            return True
        else:
            return False

    def accept_residue(self, residue):
        if residue.resname != self.residue_out and str(residue._id[0]).strip() == '':
            return True
        else:
            return False


class MMCIFfile(ProteinFile):
    def download(self):
        self.cif_file = path.join(self.dir, self.pdbcode + ".cif")
        try:
            makedirs(self.dir)
        except OSError as e:
            pass
        response = urllib.request.urlopen('http://www.ebi.ac.uk/pdbe/entry-files/download/' + self.pdbcode + '.cif')
        html = response.read().decode("UTF-8")
        with open(self.cif_file, 'w') as myfile:
            myfile.write(html)
        if self.chain is not None:
            self.filter_by_chain()
        else:
            self.out_file = self.cif_file
        return self.out_file

    def filter_by_chain(self):
        parser = MMCIFParser()
        structure = parser.get_structure(self.pdbcode, self.cif_file)
        io = MMCIFIO()
        io.set_structure(structure)
        self.out_file = self.cif_file.replace(".cif", "_" + self.chain + ".cif")
        io.save(self.out_file, ChainAndResidueSelect(self.chain))


class PdbFile(ProteinFile):
    '''
       Download PDB files from RCSB PDB Database and filter by chain
       Supports also PDB Bundles when there are many subchains for a given protein

       example:
       PdbFile("1j85", "A", "/tmp").download()

       Parameters
       ==========
       code: string - PDB ID
       chain: string
       path: string - where to store PDB file

    '''

    def download(self):
        pdbFile = self.dir + '/' + self.pdbcode + '.pdb'
        try:
            makedirs(self.dir)
        except OSError as e:
            pass
        try:
            response = urllib.request.urlopen('https://files.rcsb.org/view/' + self.pdbcode.upper() + '.pdb')
            html = response.read().decode("UTF-8")
            with open(pdbFile, 'w') as myfile:
                myfile.write(html)
            if self.chain is not None:
                pdbFile = self.filter_by_chain(pdbFile, self.chain)
        except urllib.error.HTTPError as e:
            print(self.pdbcode + " " + self.chain + " ...trying to download from PDB Bundle")
            response = urllib.request.urlopen('https://files.rcsb.org/pub/pdb/compatible/pdb_bundle/' + self.pdbcode.lower()[1:3] + '/' + self.pdbcode.lower() + '/' + self.pdbcode.lower() + '-pdb-bundle.tar.gz')
            tar = tarfile.open(fileobj=response, mode="r|gz")
            tar.extractall(self.dir)
            tar.close()
            mapFile, mapChain = self.parsePdbBundleChainIdFile(self.dir + '/' + self.pdbcode.lower() + '-chain-id-mapping.txt')
            newChain = mapChain.get(self.chain)
            pdbBundleFile = self.dir + '/' + mapFile.get(self.chain)
            if newChain!=self.chain:
                self.parsePdbAndTranslateChain(pdbBundleFile, pdbFile, self.chain, newChain)
            else:
                shutil.move(self.dir + '/' + mapFile.get(self.chain),pdbFile)
        return pdbFile

    @staticmethod
    def parsePdbBundleChainIdFile(chainFile):
        with open(chainFile, encoding='utf-8') as fp:
            line = fp.readline()
            cnt = 1
            files = []
            mapChain = {}
            mapFile = {}
            actualFile = ''
            while line:
                line = line.strip().rstrip()
                cnt += 1
                if line.find('pdb-bundle')>=0:
                    actualFile = line[:-1]
                    files.append(actualFile)
                    line = fp.readline()
                    continue
                elif actualFile!='' and line!='':
                    mapping = line.split()
                    key = mapping[1].strip()
                    val = mapping[0].strip()
                    mapChain[key]=val
                    mapFile[key] = actualFile
                line = fp.readline()
            return mapFile, mapChain

    # Remapping chain for PDB bundles with many subchains
    def parsePdbAndTranslateChain(pdbFileIn,pdbFileOut,chain,newChain):
        #print chain + '->' + newChain
        with open(pdbFileIn, "r", encoding='utf-8') as infile, open(pdbFileOut, "w", encoding='utf-8') as outfile:
            reader = csv.reader(infile)
            for i, line in enumerate(reader):
                if line[0].find('ATOM')==0 or line[0].find('HETATM')==0:
                    newLine = list(line[0])
                    if newLine[21]==newChain:
                        if len(chain) > 2:
                            newLine += chain
                            newLine[21] = "%"
                        elif len(chain)>1:
                            newLine[20] = chain[0]
                            newLine[21] = chain[1]
                        else:
                            newLine[21] = chain
                        outfile.write("".join(newLine) + "\n")
                else:
                    outfile.write(line[0] + "\n")

    @staticmethod
    def filter_by_chain(pdbfile, chain):
        pdbfile_out = pdbfile.replace(".pdb", "_" + chain + ".pdb")
        with open(pdbfile_out, "w") as outfile, open(pdbfile) as file:
            for _, line in enumerate(file):
                if line.startswith("ATOM") or line.startswith("TER"):
                    if str(line[21]) == chain:
                        outfile.write(line)
                else:
                    outfile.write(line)
        return pdbfile_out


class getIdenticalChains:
    '''
       Find identical chains to a given one

       example:
       a = getIdenticalChains("2jlo",chain="A").get()

       Parameters
       ==========
       pdbcode: string - PDB ID
       chain: string
    '''

    def __init__(self, pdbcode, chain='A'):
        self.pdb = pdbcode.upper()
        self.chain = chain
        f = urllib.request.urlopen('http://www.rcsb.org/pdb/rest/describeMol?structureId='+self.pdb+'.'+self.chain)
        data = f.read()
        f.close()
        self.root = etree.fromstring(data)

    def get(self):
        d = self.root.xpath("//molDescription/structureId[@id='"+self.pdb+"'][@chainId='"+self.chain+"']/polymer/chain/@id")
        return d


class getSimilarChains:
    '''
       Find similar chains to a given one with sequence identity given as parameter

       example:
       a = getSimilarChains("2jlo",chain="A",identity=90).get()

       Parameters
       ==========
       pdbcode: string - PDB ID
       chain: string
       identity: int - (percentage) of sequence identity
    '''
    def __init__(self, pdb, chain='A', identity=40):
        chain = chain
        url = "http://www.rcsb.org/pdb/rest/sequenceCluster?cluster="+str(identity)+"&structureId="+pdb+"."+chain
        r = urllib.request.urlopen(url)
        data = r.read()
        r.close()
        try:
            xmldoc = minidom.parseString(data)
            self.items = [elt.getAttribute('name') for elt in xmldoc.getElementsByTagName('pdbChain')]
        except:
            self.items = [pdb.upper()+"."+chain]#.upper()]

    def get(self):
        return self.items


class getUniqChains:
    '''
       Find a list of unique chains for a given PDB id

       example:
       a = getUniqChains("2jlo").get()

       Parameters
       ==========
       pdbcode: string - PDB ID
    '''
    def __init__(self, pdbcode):
        self.pdb = pdbcode.upper()
        f = urllib.request.urlopen('http://www.rcsb.org/pdb/rest/describeMol?structureId='+self.pdb)
        data = f.read()
        f.close()
        self.root = etree.fromstring(data)

    def get(self):
        o = []
        d = self.root.xpath("//molDescription/structureId[@id='"+self.pdb+"']/polymer/@entityNr")
        for e in d:
            d2 = self.root.xpath("//molDescription/structureId[@id='"+self.pdb+"']/polymer[@entityNr='"+e+"']/chain/@id")
            o.append(d2[0])
        return o


def fetchReleasedPdbs(from_date, to_date=''):
    '''
        Download list of pdb ids released in range from_date - to_date (or in day from_date). Return False if empty list
    '''
    if to_date == '':
        to_date = from_date
    '''
    date in format 2009-07-01
    '''
    url = 'http://www.rcsb.org/pdb/rest/search'
    queryText = """
<orgPdbCompositeQuery version="1.0">
<queryRefinement>
<queryRefinementLevel>0</queryRefinementLevel>
<orgPdbQuery>

<queryType>org.pdb.query.simple.ReleaseDateQuery</queryType>
<pdbx_audit_revision_history.revision_date.comparator>between</pdbx_audit_revision_history.revision_date.comparator>
<pdbx_audit_revision_history.revision_date.min>%s</pdbx_audit_revision_history.revision_date.min>
<pdbx_audit_revision_history.revision_date.max>%s</pdbx_audit_revision_history.revision_date.max>
<pdbx_audit_revision_history.ordinal.comparator>=</pdbx_audit_revision_history.ordinal.comparator>
<pdbx_audit_revision_history.ordinal.value>1</pdbx_audit_revision_history.ordinal.value>
</orgPdbQuery>
</queryRefinement>
<queryRefinement>
<queryRefinementLevel>1</queryRefinementLevel>
<conjunctionType>and</conjunctionType>
<orgPdbQuery>
<queryType>org.pdb.query.simple.ChainTypeQuery</queryType>
<description>Chain Type: there is a Protein chain</description>
<containsProtein>Y</containsProtein>
<containsDna>?</containsDna>
<containsRna>?</containsRna>
<containsHybrid>?</containsHybrid>
</orgPdbQuery>
</queryRefinement>
</orgPdbCompositeQuery>
    """ % (from_date, to_date)

    req = urllib.request.Request(url, data=queryText.encode("utf-8"), method='POST')
    f = urllib.request.urlopen(req)
    result = [i.strip().decode("utf-8") for i in f.readlines()]
    f.close()
    if result:
        return result
    else:
        return False


if __name__ == "__main__":
    a = getIdenticalChains("2jlo", chain="A").get()
    print(a)
    a = getSimilarChains("2jlo",chain="A",identity=90).get()
    print(a)
    a = getUniqChains("2jlo").get()
    print(a)
    p = PdbFile("/tmp","1k36", "A").download()
    p = MMCIFfile("/tmp", "1k36", "A").download()
    print(fetchReleasedPdbs("2017-10-01"))




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="spyprot",
    version="0.0.1.1",
    author="Borys Jastrzębski, Paweł Rubach",
    author_email="b.jastrzebski@cent.uw.edu.pl, p.rubach@cent.uw.edu.pl",
    description="Rapid protein overview from multiple databases",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Zedelghem/spyProt",
    packages=setuptools.find_packages(),
    install_requires=[
   'tqdm>=4.31.1',
   'lxml>=3.5.0',
   'requests>=2.20.0',
   'Bio>=0.1.0',
   'psutil>=5.6.3',
   'subprocess32>=3.5.4',
   'wget>=3.2'
],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)